@@ -72,14 +72,16 @@
 #define BUF_SIZE 65535
 #endif
 
-int verbose      = 0;
-char *executable = "ss-server";
-char working_dir[PATH_MAX];
+int verbose          = 0;
+char *executable     = "ss-server";
+char *working_dir    = NULL;
+int working_dir_size = 0;
 
 static struct cork_hash_table *server_table;
 
 #ifndef __MINGW32__
-static int setnonblocking(int fd)
+static int
+setnonblocking(int fd)
 {
     int flags;
     if (-1 == (flags = fcntl(fd, F_GETFL, 0))) {
@@ -90,16 +92,20 @@ static int setnonblocking(int fd)
 
 #endif
 
-static void build_config(char *prefix, struct server *server)
+static void
+build_config(char *prefix, struct server *server)
 {
-    char path[PATH_MAX];
+    char *path    = NULL;
+    int path_size = strlen(prefix) + strlen(server->port) + 20;
 
-    snprintf(path, PATH_MAX, "%s/.shadowsocks_%s.conf", prefix, server->port);
+    path = malloc(path_size);
+    snprintf(path, path_size, "%s/.shadowsocks_%s.conf", prefix, server->port);
     FILE *f = fopen(path, "w+");
     if (f == NULL) {
         if (verbose) {
             LOGE("unable to open config file");
         }
+        ss_free(path);
         return;
     }
     fprintf(f, "{\n");
@@ -107,9 +113,11 @@ static void build_config(char *prefix, struct server *server)
     fprintf(f, "\"password\":\"%s\",\n", server->password);
     fprintf(f, "}\n");
     fclose(f);
+    ss_free(path);
 }
 
-static char *construct_command_line(struct manager_ctx *manager, struct server *server)
+static char *
+construct_command_line(struct manager_ctx *manager, struct server *server)
 {
     static char cmd[BUF_SIZE];
     int i;
@@ -152,7 +160,11 @@ static char *construct_command_line(struct manager_ctx *manager, struct server *
     }
     if (manager->fast_open) {
         int len = strlen(cmd);
-        snprintf(cmd + len, BUF_SIZE - len, " --fast_open");
+        snprintf(cmd + len, BUF_SIZE - len, " --fast-open");
+    }
+    if (manager->mtu) {
+        int len = strlen(cmd);
+        snprintf(cmd + len, BUF_SIZE - len, " --mtu %d", manager->mtu);
     }
     for (i = 0; i < manager->nameserver_num; i++) {
         int len = strlen(cmd);
@@ -170,7 +182,8 @@ static char *construct_command_line(struct manager_ctx *manager, struct server *
     return cmd;
 }
 
-static char *get_data(char *buf, int len)
+static char *
+get_data(char *buf, int len)
 {
     char *data;
     int pos = 0;
@@ -185,7 +198,8 @@ static char *get_data(char *buf, int len)
     return data;
 }
 
-static char *get_action(char *buf, int len)
+static char *
+get_action(char *buf, int len)
 {
     char *action;
     int pos = 0;
@@ -204,7 +218,8 @@ static char *get_action(char *buf, int len)
     return action;
 }
 
-static struct server *get_server(char *buf, int len)
+static struct server *
+get_server(char *buf, int len)
 {
     char *data = get_data(buf, len);
     char error_buf[512];
@@ -250,7 +265,8 @@ static struct server *get_server(char *buf, int len)
     return server;
 }
 
-static int parse_traffic(char *buf, int len, char *port, uint64_t *traffic)
+static int
+parse_traffic(char *buf, int len, char *port, uint64_t *traffic)
 {
     char *data = get_data(buf, len);
     char error_buf[512];
@@ -283,7 +299,8 @@ static int parse_traffic(char *buf, int len, char *port, uint64_t *traffic)
     return 0;
 }
 
-static void add_server(struct manager_ctx *manager, struct server *server)
+static void
+add_server(struct manager_ctx *manager, struct server *server)
 {
     bool new = false;
     cork_hash_table_put(server_table, (void *)server->port, (void *)server, &new, NULL, NULL);
@@ -294,16 +311,19 @@ static void add_server(struct manager_ctx *manager, struct server *server)
     }
 }
 
-static void kill_server(char *prefix, char *pid_file)
+static void
+kill_server(char *prefix, char *pid_file)
 {
-    char path[PATH_MAX];
-    int pid;
-    snprintf(path, PATH_MAX, "%s/%s", prefix, pid_file);
+    char *path = NULL;
+    int pid, path_size = strlen(prefix) + strlen(pid_file) + 2;
+    path = malloc(path_size);
+    snprintf(path, path_size, "%s/%s", prefix, pid_file);
     FILE *f = fopen(path, "r");
     if (f == NULL) {
         if (verbose) {
             LOGE("unable to open pid file");
         }
+        ss_free(path);
         return;
     }
     if (fscanf(f, "%d", &pid) != EOF) {
@@ -311,27 +331,33 @@ static void kill_server(char *prefix, char *pid_file)
     }
     fclose(f);
     remove(path);
+    ss_free(path);
 }
 
-static void stop_server(char *prefix, char *port)
+static void
+stop_server(char *prefix, char *port)
 {
-    char path[PATH_MAX];
-    int pid;
-    snprintf(path, PATH_MAX, "%s/.shadowsocks_%s.pid", prefix, port);
+    char *path = NULL;
+    int pid, path_size = strlen(prefix) + strlen(port) + 20;
+    path = malloc(path_size);
+    snprintf(path, path_size, "%s/.shadowsocks_%s.pid", prefix, port);
     FILE *f = fopen(path, "r");
     if (f == NULL) {
         if (verbose) {
             LOGE("unable to open pid file");
         }
+        ss_free(path);
         return;
     }
     if (fscanf(f, "%d", &pid) != EOF) {
         kill(pid, SIGTERM);
     }
     fclose(f);
+    ss_free(path);
 }
 
-static void remove_server(char *prefix, char *port)
+static void
+remove_server(char *prefix, char *port)
 {
     char *old_port            = NULL;
     struct server *old_server = NULL;
@@ -345,7 +371,8 @@ static void remove_server(char *prefix, char *port)
     stop_server(prefix, port);
 }
 
-static void update_stat(char *port, uint64_t traffic)
+static void
+update_stat(char *port, uint64_t traffic)
 {
     void *ret = cork_hash_table_get(server_table, (void *)port);
     if (ret != NULL) {
@@ -354,7 +381,8 @@ static void update_stat(char *port, uint64_t traffic)
     }
 }
 
-static void manager_recv_cb(EV_P_ ev_io *w, int revents)
+static void
+manager_recv_cb(EV_P_ ev_io *w, int revents)
 {
     struct manager_ctx *manager = (struct manager_ctx *)w;
     socklen_t len;
@@ -476,7 +504,8 @@ ERROR_MSG:
     }
 }
 
-static void signal_cb(EV_P_ ev_signal *w, int revents)
+static void
+signal_cb(EV_P_ ev_signal *w, int revents)
 {
     if (revents & EV_SIGNAL) {
         switch (w->signum) {
@@ -487,7 +516,8 @@ static void signal_cb(EV_P_ ev_signal *w, int revents)
     }
 }
 
-int create_server_socket(const char *host, const char *port)
+int
+create_server_socket(const char *host, const char *port)
 {
     struct addrinfo hints;
     struct addrinfo *result, *rp, *ipv4v6bindall;
@@ -561,7 +591,8 @@ int create_server_socket(const char *host, const char *port)
     return server_sock;
 }
 
-int main(int argc, char **argv)
+int
+main(int argc, char **argv)
 {
     int i, c;
     int pid_flags         = 0;
@@ -578,6 +609,7 @@ int main(int argc, char **argv)
     int auth      = 0;
     int fast_open = 0;
     int mode      = TCP_ONLY;
+    int mtu       = 0;
 
     int server_num = 0;
     char *server_host[MAX_REMOTE_NUM];
@@ -589,11 +621,12 @@ int main(int argc, char **argv)
 
     int option_index                    = 0;
     static struct option long_options[] = {
-        { "fast-open"      , no_argument      , 0, 0 },
-        { "acl"            , required_argument, 0, 0 },
+        { "fast-open",       no_argument,       0, 0 },
+        { "acl",             required_argument, 0, 0 },
         { "manager-address", required_argument, 0, 0 },
-        { "executable"     , required_argument, 0, 0 },
-        { "help"           , no_argument      , 0, 0 },
+        { "executable",      required_argument, 0, 0 },
+        { "mtu",             required_argument, 0, 0 },
+        { "help",            no_argument,       0, 0 },
         {                 0,                 0, 0, 0 }
     };
 
@@ -614,6 +647,9 @@ int main(int argc, char **argv)
             } else if (option_index == 3) {
                 executable = optarg;
             } else if (option_index == 4) {
+                mtu = atoi(optarg);
+                LOGI("set MTU to %d", mtu);
+            } else if (option_index == 5) {
                 usage();
                 exit(EXIT_SUCCESS);
             }
@@ -667,6 +703,7 @@ int main(int argc, char **argv)
             break;
         case '?':
             // The option character is not recognized.
+            LOGE("Unrecognized option: %s", optarg);
             opterr = 1;
             break;
         }
@@ -706,6 +743,9 @@ int main(int argc, char **argv)
         if (mode == TCP_ONLY) {
             mode = conf->mode;
         }
+        if (mtu == 0) {
+            mtu = conf->mtu;
+        }
     }
 
     if (server_num == 0) {
@@ -776,8 +816,9 @@ int main(int argc, char **argv)
     manager.host_num        = server_num;
     manager.nameservers     = nameservers;
     manager.nameserver_num  = nameserver_num;
+    manager.mtu             = mtu;
 
-    // inilitialize ev loop
+    // initialize ev loop
     struct ev_loop *loop = EV_DEFAULT;
 
     // setuid
@@ -787,11 +828,14 @@ int main(int argc, char **argv)
 
     struct passwd *pw   = getpwuid(getuid());
     const char *homedir = pw->pw_dir;
-    snprintf(working_dir, PATH_MAX, "%s/.shadowsocks", homedir);
+    working_dir_size = strlen(homedir) + 15;
+    working_dir      = malloc(working_dir_size);
+    snprintf(working_dir, working_dir_size, "%s/.shadowsocks", homedir);
 
     int err = mkdir(working_dir, S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH);
     if (err != 0 && errno != EEXIST) {
         ERROR("mkdir");
+        ss_free(working_dir);
         FATAL("unable to create working directory");
     }
 
@@ -804,11 +848,13 @@ int main(int argc, char **argv)
             size_t len = strlen(ep->d_name);
             if (strcmp(ep->d_name + len - 3, "pid") == 0) {
                 kill_server(working_dir, ep->d_name);
-                if (verbose) LOGI("kill %s", ep->d_name);
+                if (verbose)
+                    LOGI("kill %s", ep->d_name);
             }
         }
-        closedir (dp);
+        closedir(dp);
     } else {
+        ss_free(working_dir);
         FATAL("Couldn't open the directory");
     }
 
@@ -831,6 +877,7 @@ int main(int argc, char **argv)
         struct sockaddr_un svaddr;
         sfd = socket(AF_UNIX, SOCK_DGRAM, 0);       /*  Create server socket */
         if (sfd == -1) {
+            ss_free(working_dir);
             FATAL("socket");
         }
 
@@ -838,6 +885,7 @@ int main(int argc, char **argv)
 
         if (remove(manager_address) == -1 && errno != ENOENT) {
             ERROR("bind");
+            ss_free(working_dir);
             exit(EXIT_FAILURE);
         }
 
@@ -847,11 +895,13 @@ int main(int argc, char **argv)
 
         if (bind(sfd, (struct sockaddr *)&svaddr, sizeof(struct sockaddr_un)) == -1) {
             ERROR("bind");
+            ss_free(working_dir);
             exit(EXIT_FAILURE);
         }
     } else {
         sfd = create_server_socket(ip_addr.host, ip_addr.port);
         if (sfd == -1) {
+            ss_free(working_dir);
             FATAL("socket");
         }
     }
@@ -884,6 +934,7 @@ int main(int argc, char **argv)
 
     ev_signal_stop(EV_DEFAULT, &sigint_watcher);
     ev_signal_stop(EV_DEFAULT, &sigterm_watcher);
+    ss_free(working_dir);
 
     return 0;
 }
